<?php

namespace Scriptor\Modules\Rheostatic;

use Imanager\Util;
use GuzzleHttp\Client;
use Scriptor\Core\Helper;
use Scriptor\Core\Module;
use Scriptor\Core\Scriptor;
use Scriptor\Core\Site;

/**
 * Rheostatic Module Class
 *
 * This Scriptor module is responsible for 
 * the generation of the static pages.
 *
 */
class Rheostatic extends Module
{
	/**
	 * Module version 
	 * 
	 * Use ~~ echo Rheostatic::VERSION; ~~~
	 * to get the current module version
	 */
	const VERSION = '1.5.1';

	/**
	 * Lacal config variable
	 * see custom.scriptor-config.php file 
	 * 'modules' => 'rheostatic' section.
	 * 
	 * @var array
	 */
	public $locConfig;

	/**
	 * Path where Rheostatic stores the static pages, 
	 * you can define it in the config.
	 * 
	 * @var string
	 */
	private $dataPath;

	/**
	 * GuzzleHttp\Client instance
	 */
	private $gClient;

	/**
	 * Imanager\Item instance
	 */
	private static $pageClone;

	/**
	 * Pages category
	 */
	public $pages;


	private static $initialized = false;

	public static function moduleInfo() : array
	{
		return [
			'name' => 'Rheostatic',
			'menu' => 'rheostatic_menu',
            'position' => 3,
            'active' => true,
            'auth' => true,
            'autoinit' => true,
			'display_type' => [
				'sidebar'
			],
			'icon' => 'gg-list-tree',
            'description' => 'Scriptor\'s Static Site Generator: Transform your dynamic Scriptor CMS website into a fast, secure, and portable static site. <a href="https://scriptor-cms.info/extensions/extensions-modules/rheostatic/">More information</a>',
			'version' => self::VERSION,
			'author' => 'Bigin',
			'author_website' => 'https://ehret-studio.com',
			'date_time_format' => 'Y.m.d H:i',
			'client' => [
				'verify' => true,
				'base_uri' => '',
				'request' => [
					'query' => []
				]
			],
			'replacement' => []
		];
	}

	public static function moduleHooks() : array
	{
		return [
			'Editor::beforeExecute' => [
				[
					'module' => 'Rheostatic', 
					'method' => 'loadResources'
				]
			],
			'Pages::beforeRemovePage' => [
				[
					'module' => 'Rheostatic', 
					'method' => 'clonePage'
				]
			],
			'Pages::afterRemovePage' => [
				[
					'module' => 'Rheostatic', 
					'method' => 'removePageFromIndex'
				]
			]
		];
	}

	/**
	 * Initializes Rheostatic module
	 */
	public function init()
	{
		if (!self::$initialized) {
			parent::init();
			$this->locConfig = $this->config['modules']['Rheostatic'];
			Scriptor::load(dirname(__DIR__, 3).'/'.$this->config['admin_path'].'/core/site.php');
			$this->site = new Site();
			$this->site->init();
			Scriptor::load(__DIR__.'/guzzle/autoload.php');
			$this->csrf = Scriptor::getCSRF();
			$options = [];
			if (isset($this->locConfig['client']['verify'])) {
				$options['verify'] = $this->locConfig['client']['verify'];
			}
			$this->gClient = new Client($options);
			self::$initialized = true;
		}
	}

	/**
	 * Is executed when Rheostatic page is requested 
	 * in the Scriptor editor.
	 */
	public function execute()
	{
		$this->checkAction();

		if ($this->segments->get(0) == 'Rheostatic' && !$this->segments->get(1)) {
			$this->pageTitle = 'Rheostatic Pages List';
			$this->pageContent = $this->renderPageList();
			$this->breadcrumbs = '<li><span>'.$this->i18n['rheostatic_menu'].'</span></li>';
		}
		elseif ($this->segments->get(0) == 'Rheostatic' && $this->segments->get(1) == 'add') {
			$this->createStaticPage((int)$this->input->get->page);
			Util::redirect('../');
		}
		elseif ($this->segments->get(0) == 'Rheostatic' && $this->segments->get(1) == 'remove') {
			$this->removeStaticPage((int)$this->input->get->page);
			Util::redirect('../');
		}
	}

	/**
	 * There is only one action on page save
	 */
	protected function checkAction() 
	{
		// Save page data
		if ($this->input->post->action == 'add-page') {
			$this->addPage();
		}
	}

	/**
	 * Renders pagelist view in the editor
	 */
	private function renderPageList()
	{
		$rows = '<tr><td colspan="5">'.$this->i18n['no_page'].'</td></tr>';
		if ($this->site->pages()->category->items) {
			$rows = $this->renderRows();
		}
		ob_start(); ?>
		<div id="page-list-wrapper">
			<h1><?php echo $this->i18n['pages_header'] ?></h1>
			<form id="page-list-form" action="./" method="post">
				<table id="page-list-table">
					<thead>
					<tr>
						<th><b><?php echo $this->i18n['id_table_header']; ?></b></th>
						<th><b><?php echo $this->i18n['parent_table_header']; ?></b></th>
						<th><b><?php echo $this->i18n['title_table_header']; ?></b></th>
						<th><b><?php echo $this->i18n['created_table_header']; ?></b></th>
						<th><b><?php echo $this->i18n['action_table_header']; ?></b></th>
					</tr>
					</thead>
					<tbody>
					<?php echo $rows; ?>
					</tbody>
				</table>
			</form>
		</div>
		<a class="btn" href="#"><button class="icons button" name="button" value="<?php 
			echo $this->i18n['add_pages_button']; ?>" type="button">
			<i class="gg-stack"></i>&nbsp;<?php 
			echo $this->i18n['add_pages_button']; ?></button></a>
		<script>
			var loaderText = '<?php echo $this->i18n['loader_start_text'] ?>';
		</script>
		<script src="<?php echo dirname($this->siteUrl).'/site/modules/Rheostatic/scripts/rheostatic.js' ?>"></script>
		<?php
		return ob_get_clean();
	}

	protected function renderRows()
	{
		$rows = '';
		$sorted = $this->site->pages()->category->sort('position', 'asc',  0, 0, $this->site->pages()->category->items);
		$token = $this->csrf->renderUrl('&');
		
		foreach($sorted as $page) {	
			$url = '';
			if ($page->id != 1) $url = Site::getPageUrl($page, $this->site->pages);
			if(!$this->dataPath) $this->dataPath = $this->getStaticPath();
			
			$date = '';
			$remove = '';
			if(file_exists(dirname($this->dataPath)."/{$url}index.html")) {
				$date = date($this->locConfig['date_time_format'], filemtime(dirname($this->dataPath)."/{$url}index.html"));
				$remove = '<a class="add i-wrapp-inline dfn-hover remove" rel="'.$this->i18n['pre_remove_static_page_msg']
					.'" href="remove/?page='.$page->id.$token.'"><dfn data-info="'.$this->i18n['remove_page_tooltip']
					.'"><i class="gg-duplicate-remove"></i></dfn></a>';
			} else {
				$remove = '<span class="add i-wrapp-inline dfn-hover"><i class="gg-duplicate-remove"></i></span>';
			}

			$rows .= '
				<tr>
					<td>'.$page->id.'</td>
					<td>'.(($page->parent) ? $page->parent : '').'</td>
					<td><a href="../pages/edit/?page='.$page->id.'">'.((mb_strlen($page->name) > 80) ? 
						mb_substr($page->name, 0,80).'...' : $page->name).'</a></td>
					<td id="cdate-'.$page->id.'">'.$date.'</td>
					<td><a class="add i-wrapp-inline dfn-hover" href="add/?page='.$page->id.$token.'"><dfn 
						data-info="'.$this->i18n['add_page_tooltip'].'"><i class="gg-duplicate"></i></dfn></a>
						'.$remove.'
						</td>
				</tr>
			';
		}
		return $rows;
	}

	private function addPage() 
	{
		if(!$this->input->post->page) { return false; }
		$pageId = (int) $this->input->post->page;

		$pages = $this->site->pages()->category->items;
		$item = $pages[$pageId];
		$next = null;

		if($item) {
			$next = $this->getNext($pages, $pageId);
			if($this->createStaticPage($item->id)) {
				array_pop($this->msgs);
				$cDate = date($this->locConfig['date_time_format']);
				sleep(1);
				if($next) {
					$this->jsonResponse(['status' => 1, 
						'nextId' => $next->id, 
						'loaderText' => 'The page '.$next->id.' in progress...', 
						'cDate' => $cDate
					]);
				}
				$this->msgs[] = array(
					'type' => 'success',
					'value' => Helper::rawClassName(__CLASS__).': '.
						$this->i18n['pages_successful_added']
				);
				$this->jsonResponse(['status' => 1, 'redirect' => './']);
			}
		}
		$this->jsonResponse(['status' => 0, 'redirect' => './']);
	}

	private function createStaticPage(int $page_id)
	{
		$page = $this->site->pages()->getPage($page_id);

		if(!$page) {
			$this->msgs[] = array(
				'type' => 'error',
				'value' => Helper::rawClassName(__CLASS__).': '.
					$this->i18n['error_page_not_found']
			);
			return false;
		}
		/* if($this->config['protectCSRF'] && !$this->csrf->isTokenValid(
			$this->input->get->tokenName,
			$this->input->get->tokenValue, true)) {
				$this->msgs[] = array(
					'type' => 'error',
					'value' => Helper::rawClassName(__CLASS__).
						': '.$this->i18n['error_csrf_token_mismatch']
				);
			return false;
		}*/
		$url = '';
		if($page->id != 1) $url = Site::getPageUrl($page, $this->site->pages);

		if(!$this->dataPath) $this->dataPath = $this->getStaticPath();

		$path = dirname($this->dataPath)."/{$url}index.html";
		//Util::install($path);
		$baseUrl = dirname($this->siteUrl);
		if(isset($this->locConfig['client']['base_uri']) && !empty($this->locConfig['client']['base_uri'])) {
			$baseUrl = $this->locConfig['client']['base_uri'];
		}
		if(isset($this->locConfig['client']['request']) && !empty($this->locConfig['client']['request'])) {
			$response = $this->gClient->get($baseUrl.'/'.$url, $this->locConfig['client']['request']);
		} else {
			$response = $this->gClient->get($baseUrl.'/'.$url);
		}
		$code = $response->getStatusCode();

		if($code == 200) {
			Util::install($path);
			$body = $response->getBody();

			if(isset($this->locConfig['replacement']) && !empty($this->locConfig['replacement'])) {
				foreach($this->locConfig['replacement'] as $key => $repl) {
					$body = str_replace($key, $repl, $body);
				}
			}
			if(false === file_put_contents($path, $body)) {
				return false;
			}

			$this->msgs[] = array(
				'type' => 'success',
				'value' => Helper::rawClassName(__CLASS__).
					': '.$this->i18n['page_successful_added']
			);
			return true;
		}
		elseif(isset($this->i18n["status_$code"])) {
			$this->msgs[] = array(
				'type' => 'error',
				'value' => Helper::rawClassName(__CLASS__).
					': '.$this->i18n["status_$code"]
			);
			return false;
		}

		$this->msgs[] = array(
			'type' => 'error',
			'value' => Helper::rawClassName(__CLASS__).
				': '.$code.$this->i18n['status_code']
		);
		return false;
	}

	private function removeStaticPage(int $page_id)
	{
		// First, check if there is a cloned of that page
		if(self::$pageClone && self::$pageClone->id == $page_id) {
			$page = self::$pageClone;
		} else {
			$page = $this->site->pages->getPage($page_id);
		}
		
		if(!$page) {
			$this->msgs[] = array(
				'type' => 'error',
				'value' => Helper::rawClassName(__CLASS__).
					': '.$this->i18n['error_page_not_found']
			);
			return false;
		}

		/* if($this->config['protectCSRF'] && !$this->csrf->isTokenValid(
			$this->input->get->tokenName,
			$this->input->get->tokenValue, true)) {
				$this->msgs[] = array(
					'type' => 'error',
					'value' => Helper::rawClassName(__CLASS__).
						': '.$this->i18n['error_csrf_token_mismatch']
				);
			return false;
		}*/
		$url = '';
		if($page->id != 1) $url = Site::getPageUrl($page, $this->site->pages);
		if(!$this->dataPath) $this->dataPath = $this->getStaticPath();

		$path = dirname($this->dataPath)."/{$url}index.html";

		if(file_exists($path)) {
			if(Util::delTree(dirname($path))) {
				// If the root directory is empty, delete it
				/* if($page->id == 1 && $this->isDirEmpty(dirname($this->dataPath))) {
					Util::delTree(dirname($this->dataPath));
				} */
				$this->msgs[] = array(
					'type' => 'success',
					'value' => Helper::rawClassName(__CLASS__).
						': '.$this->i18n['static_page_successful_removed']
				);
				return true;
			}
		}

		$this->msgs[] = array(
			'type' => 'error',
			'value' => Helper::rawClassName(__CLASS__).
				': '.$this->i18n['error_removing_static_page']
		);
		return false;
	}

	/**
	 * Inserts Reostatic style sheet to the head section of Scriptor editor.
	 */
	public function loadResources()
	{
		$this->addResource(
			'link', ['rel' => 'stylesheet',
				'href' => dirname($this->siteUrl).
				'/site/modules/Rheostatic/css/styles.css'], 'header'
		);
	}

	/**
	 * This is used to clone the current page.
	 *  
	 * After deleting the page or modifying its slug, the 
	 * data would not match the static page and the static 
	 * page would not be found.
	 * 
	 */
	public function clonePage()
	{
		$page = $this->site->pages->getPage((int) $this->input->get->page);
		if ($page) {
			self::$pageClone = clone $page;
		}
	}

	public function removePageFromIndex($event)
	{
		if(!$event->return) return;

		$id = isset($event->object->page->id) ? 
			$event->object->page->id : $this->input->get->page;

		if(!$this->removeStaticPage((int) $id)) {
			array_pop($this->msgs);
		}
	}

	/**
	 * Returns the path where the static page should be stored.
	 * 
	 * Note that "installed/" at the end can be confusing, but 
	 * is needed for the function that will generate the 
	 * directory. It can also be a differ suffix, it is cut out
	 * by the function anyway.
	 * 
	 * See: Imanager\Util::install($path)
	 * 
	 * ~~ mkdir(dirname($path) ... ~~
	 */
	private function getStaticPath()
	{
		return (isset($this->locConfig['data_path']) && !empty($this->locConfig['data_path'])) ? 
				rtrim($this->locConfig['data_path'], '/').'/installed' : IM_ROOTPATH.'static/installed';
	}

	private function jsonResponse($data)
	{
		header('Content-type: application/json; charset=utf-8');
		echo json_encode($data);
		exit;
	}

	private function getNext($array, $key) 
	{
		$currentKey = key($array);
		while($currentKey !== null && $currentKey != $key) {
			next($array);
			$currentKey = key($array);
		}
		return next($array);
	}

	private function isDirEmpty($dir) 
	{
		if(!is_readable($dir)) return null;

		$handle = opendir($dir);
		while(false !== ($entry = readdir($handle))) {
		  if($entry != '.' && $entry != '..') return false;
		}
		return true;
	}
}